#include <stdlib.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <string.h>
#include <math.h>
#include "em_mog.h"
#include "macros.h"

static void
printf_matrix(const double *m, int ncol, int nrow)
{
	const double *ptr = m;
	int i, j;
	printf("\n");
	for (i = 0; i < nrow; i++) {
		for (j = 0; j < ncol; j++) {
			printf("%f ", *ptr);
			ptr++;
		}
		printf("\n");
	}
}

/*
 * em_cholesky_decomp: cholesky decomposition from matrix with dimension 3x3
 */
static int
em_cholesky_decomp_3x3(double *C)
{
	return_val_if_fail(C->size1 == 3 && C->size2 == 3, EM_MOG_ERROR);
	double *d = C;
	double tmp;
	if (*d <= 0)
		goto err;
	/* L[1,1] = sqrt(A[1,1]) */
	*d = sqrt(*d);
	/* L[2,1] = 1 / L[1,1] * (A[2,1])*/
	d[3] /= d[0];
	tmp = d[4] - d[3] * d[3];
	if (tmp <= 0.0)
		goto err;
	d[4] = sqrt(tmp);

	d[6] /= d[0];
	d[7] = (d[7] - d[3] * d[6]) / d[4];
	tmp = d[8] - d[6] * d[6] - d[7] * d[7];
	if (tmp <= 0.0)
		goto err;
	if (tmp <= 0.0)
			goto err;
	d[8] = sqrt(tmp);
	goto quit;

err:
	fprintf(stderr, "em_cholesky_decomp: matrix must be positive definite\n");
	printf_matrix(d, 3, 3);
	return EM_MOG_ERROR;
	exit(0);
quit:;
	return EM_MOG_SUCCESS;
}

/**
 * @function
 * return normal distribution for 3 dimension data
 * matrix cmx must be diagonal
 */
inline static double
normal_distribution_diag_cov_mtrx(const double *mean, const double *x, const double *cmx)
{
	return_val_if_fail(mean != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(cmx != NULL, EM_MOG_ERROR);
	return_val_if_fail(dim > 0, EM_MOG_ERROR);
	double tmp, degree = 0.0, ret = 0.0, det = 1.0, log_exp;

	const double *m, *v, *p;
	m = cmx;
	v = mean;
	p = x;

	det *= *m;
	tmp = (*v - *p);
	degree += tmp * tmp / *m;
	m += 4;
	v++;
	p++;

	det *= *m;
	tmp = (*v - *p);
	degree += tmp * tmp / *m;
	m += 4;
	v++;
	p++;

	det *= *m;
	tmp = (*v - *p);
	degree += tmp * tmp / *m;
	log_exp = -0.5 * (M3_EM_MOG_LOG_2_PI + degree + log(det));

	ret = exp(log_exp);

	return ret;
}

/**
 * @function
 * return normal distribution for 3 dimension data
 * uses the Cholesky decomposition to speed up the computation
 */
static double
normal_distribution(const double *mean, const double *x, const double *cmx)
{
	return_val_if_fail(mean != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(cmx != NULL, EM_MOG_ERROR);
	return_val_if_fail(dim > 0, EM_MOG_ERROR);
	double vect[3] = {0, 0, 0};
	double rgb_cmx[9];
	double tmp, degree = 0.0, ret = 0.0, det = 1.0, log_exp;
	int res, cmx_size, dim;
	dim = 3;
	cmx_size = 9;

	memcpy(rgb_cmx, cmx, cmx_size * sizeof(double));

	res = em_cholesky_decomp_3x3(rgb_cmx);

	return_if_fail(res != EM_MOG_SUCCESS);

	double *m = rgb_cmx;
	/** the solution of linear equations by Gauss for the triangular matrix
	 * |a11   0   0|
	 * |a21 a22   0|
	 * |a31 a32 a33|
	 */
	/*first row*/
	tmp = x[0] - mean[0];
	vect[0] = tmp / *m;
	det = *m * det;
	degree += vect[0] * vect[0];
	m += 3;

	/*second row*/
	tmp = x[1] - mean[1];
	tmp -= *m * vect[0];
	m++;
	vect[1] = tmp / *m;
	det = *m * det;
	degree += vect[1] * vect[1];
	m += 2;

	/*third row*/
	tmp = x[2] - mean[2];
	tmp -= *m * vect[0];
	m++;
	tmp -= *m * vect[1];
	m++;
	vect[2] = tmp / *m;
	det = *m * det;
	degree += vect[2] * vect[2];

	log_exp = -0.5 * (M3_EM_MOG_LOG_2_PI + degree) - log(det);

	ret = exp(log_exp);

	return ret;
}

static double
log_exp(const double *mean, const double *x, const double *cmx)
{
	return_val_if_fail(mean != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(cmx != NULL, EM_MOG_ERROR);
	return_val_if_fail(dim > 0, EM_MOG_ERROR);
	double vect[3] = {0, 0, 0};
	double rgb_cmx[9];
	double tmp, degree = 0.0, det = 1.0, log_exp;
	int res;

	memcpy(rgb_cmx, cmx, 9 * sizeof(double));

	res = em_cholesky_decomp_3x3(rgb_cmx);

	return_if_fail(res != EM_MOG_SUCCESS);

	double *m = rgb_cmx;
	/** the solution of linear equations by Gauss for the triangular matrix
	 * |a11   0   0|
	 * |a21 a22   0|
	 * |a31 a32 a33|
	 */
	/*first row*/
	tmp = x[0] - mean[0];
	vect[0] = tmp / *m;
	det = *m * det;
	degree += vect[0] * vect[0];
	m += 3;

	/*second row*/
	tmp = x[1] - mean[1];
	tmp -= *m * vect[0];
	m++;
	vect[1] = tmp / *m;
	det = *m * det;
	degree += vect[1] * vect[1];
	m += 2;

	/*third row*/
	tmp = x[2] - mean[2];
	tmp -= *m * vect[0];
	m++;
	tmp -= *m * vect[1];
	m++;
	vect[2] = tmp / *m;
	det = *m * det;
	degree += vect[2] * vect[2];

	log_exp = -0.5 * (M3_EM_MOG_LOG_2_PI + degree + log(det));
	return log_exp;
}

/**
 * @function
 * return logarithm normal distribution for 3 dimension data
 * matrix cmx must be diagonal
 */
inline static double
log_exp_diag_cov_mtrx(const double *mean, const double *x, const double *cmx)
{
	return_val_if_fail(mean != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(cmx != NULL, EM_MOG_ERROR);
	return_val_if_fail(dim > 0, EM_MOG_ERROR);
	double tmp, degree = 0.0, det = 1.0, log_exp;
    const double *m, *v, *p;
	m = cmx;
	v = mean;
	p = x;

	det *= *m;
	tmp = (*v - *p);
	degree += tmp * tmp / *m;
	m += 4;
	v++;
	p++;

	det *= *m;
	tmp = (*v - *p);
	degree += tmp * tmp / *m;
	m += 4;
	v++;
	p++;

	det *= *m;
	tmp = (*v - *p);
	degree += tmp * tmp / *m;

	log_exp = -0.5 * (M3_EM_MOG_LOG_2_PI + degree + log(det));


	return log_exp;
}

/**
 * @function
 * EM algorithm expectation step
 */
static int
expectation_step(double *probability, const struct em_mog_context *em_ctx,
		         const double *data, int data_size)
{
	return_val_if_fail(probability, EM_MOG_ERROR);
	return_val_if_fail(em_ctx, EM_MOG_ERROR);
	return_val_if_fail(data, EM_MOG_ERROR);
	return_val_if_fail(em_ctx->means, EM_MOG_ERROR);

	double denum, *p1, *p2, *m, *cov_mtx, *w;
	const double *x;
	int i, j, n, dim, cov_size;
	n = em_ctx->ngauss;
	dim      = em_ctx->dim;
	cov_size = 9;
	p1 = p2 = probability;
	x = data;

	for (i = 0; i < data_size; i++) {
		denum = 0.0;
		m       = em_ctx->means;
		cov_mtx = em_ctx->cov_matrix;
		w       = em_ctx->weight;
		for (j = 0; j < em_ctx->ngauss; j++) {
			*p1 = *w * normal_distribution_diag_cov_mtrx(m, x, cov_mtx);
			denum += *p1;
			m += dim;
			cov_mtx += cov_size;
			p1++;
			w++;
		}

		for (j = 0; j < em_ctx->ngauss; j++) {
			if (denum > 0)
				*p2 /= denum;
			p2++;
		}
		x += dim;
	}

	return EM_MOG_SUCCESS;
}

/**
 * @function
 * adds to the original matrix a square matrix of vectors multiplied by the factor
 */
static void
add_vector_sqr_to_matrix(double *matrix, const double *vector, double mult)
{
	return_if_fail(matrix != NULL);
	return_if_fail(vector != NULL);

	double *m;
	const double *p1;

	m = matrix;
	p1 = vector;

	*m += mult * (*p1) * (*p1);
	m += 4;
	p1++;

	*m += mult * (*p1) * (*p1);
	m += 4;
	p1++;

	*m += mult * (*p1) * (*p1);
}

/**
 * @function
 * add vector sourse multiplied by the factor to vector dest
 */
static void
add_vector_to_vector(double *dest, const double *sourse, double mult)
{
	double *p1;
	const double *p2;
	p1 = dest;
	p2 = sourse;
	*p1 += *p2 * mult;
	p1++;
	p2++;
	*p1 += *p2 * mult;
	p1++;
	p2++;
	*p1 += *p2 * mult;
}

/**
 * @function
 * EM algorithm maximization step
 */
static int
maximization_step(const double *probability, struct em_mog_context *em_ctx,
		          const double *data, int data_size)
{
	return_val_if_fail(probability, EM_MOG_ERROR);
	return_val_if_fail(em_ctx, EM_MOG_ERROR);
	return_val_if_fail(data, EM_MOG_ERROR);
	return_val_if_fail(em_ctx->means, EM_MOG_ERROR);

	double *w, *m, *cmx, vector[3], cmx_add;
	const double *p, *x;
	int i, j, k, dim, cov_size, ngauss;

	w = em_ctx->weight;
	m = em_ctx->means;
	cmx = em_ctx->cov_matrix;
	cmx_add = em_ctx->noise_level;
	ngauss = em_ctx->ngauss;
	dim = 3;
	cov_size = 9;

	memset(w, 0, ngauss * sizeof(double));
	memset(m, 0, ngauss * dim * sizeof(double));
	memset(cmx, 0, ngauss * cov_size * sizeof(double));

	for (i = 0; i < ngauss; i++) {
		p = probability + i;
		for (j = 0; j < data_size; j++) {
			*w += *p;
			p += ngauss;
		}
		if (*w == 0.0) continue;
		p = probability + i;
		for (j = 0, x = data; j < data_size; j++) {
			double mult = *p;
			add_vector_to_vector(m, x, mult);
			p += ngauss;
			x += dim;
		}

		m[0] /= *w;
		m[1] /= *w;
		m[2] /= *w;

		p = probability + i;
		for (j = 0, x = data; j < data_size; j++) {
			for (k = 0; k < dim; k++)
				vector[k] = x[k] - m[k];
			double mult = *p;
			add_vector_sqr_to_matrix(cmx, vector, mult);
			p += ngauss;
			x += dim;
		}
		cmx[0] /= *w;
		cmx[4] /= *w;
		cmx[8] /= *w;
		cmx[0] += cmx_add;
		cmx[4] += cmx_add;
		cmx[8] += cmx_add;

		*w /= data_size;
		w++;
		m += dim;
		cmx += cov_size;
	}

	return EM_MOG_SUCCESS;
}

/**
 * @function
 * return em_mog_context likelihood
 */
static double
em_likelihood(struct em_mog_context *em_ctx, const double *data, int data_size)
{
	double *log_vect;;
	const double *x, *w, *m, *cmx;

	double sum_exp, loglike = 0;
	int i, j, dim, cov_size;
	dim = em_ctx->dim;
	cov_size = dim * dim;
	em_ctx->min_loglike = 1.0;

	log_vect = malloc(em_ctx->ngauss * sizeof(double));

	x = data;

	for (i = 0; i < data_size; i++) {
		w = em_ctx->weight;
		m = em_ctx->means;
		cmx = em_ctx->cov_matrix;
		sum_exp = 0;

		double tmp = log(*w) + log_exp_diag_cov_mtrx(m, x, cmx);
		log_vect[0] = tmp;
		w++;
		m   += dim;
		cmx += cov_size;

		for (j = 1; j < em_ctx->ngauss; j++) {
			log_vect[j] = log(*w) + log_exp_diag_cov_mtrx(m, x, cmx);

			w++;
			m   += dim;
			cmx += cov_size;

			if (log_vect[j] > tmp)
				tmp = log_vect[j];
		}

		for (j = 0; j < em_ctx->ngauss; j++)
			sum_exp += exp(log_vect[j] - tmp);

		tmp += sum_exp;

		if (tmp < em_ctx->min_loglike)
			em_ctx->min_loglike = tmp;

		loglike += tmp;
		x += dim;
	}

	free(log_vect);
	return loglike;
}

/**
 * @function
 *
 */
static double
em_data_probability(const struct em_mog_context *em_ctx, const double *data)
{
	double *log_vect;
	const double *x, *w, *m, *cmx;

	double sum_exp, loglike = 0;
	int j, dim, cov_size;
	dim = em_ctx->dim;
	cov_size = dim * dim;

	log_vect = malloc(em_ctx->ngauss * sizeof(double));

	x = data;
	w = em_ctx->weight;
	m = em_ctx->means;
	cmx = em_ctx->cov_matrix;
	sum_exp = 0;

	double tmp = log(*w) + log_exp(m, x, cmx);
	log_vect[0] = tmp;

	for (j = 1; j < em_ctx->ngauss; j++) {
		w++;
		m   += dim;
		cmx += cov_size;

		log_vect[j] = log(*w) + log_exp(m, x, cmx);

		if (log_vect[j] > tmp)
			tmp = log_vect[j];
	}

	for (j = 0; j < em_ctx->ngauss; j++)
		sum_exp += exp(log_vect[j] - tmp);

	loglike += tmp + sum_exp;

	free(log_vect);
	return loglike;
}

struct em_mog_context *
em_mog_new()
{
	struct em_mog_context *ptr;
	ptr = malloc(sizeof(struct em_mog_context));
	return_if_fail(ptr != NULL);
	ptr->cov_matrix  = NULL;
	ptr->means       = NULL;
	ptr->weight      = NULL;
	ptr->dim         = 3;
	ptr->ngauss      = 0;
	ptr->noise_level = 0.0;
	return ptr;
}

void
em_mog_free(struct em_mog_context **ctx)
{
	return_if_fail(ctk == NULL);
	return_if_fail((*ctx) == NULL);
	if ((*ctx)->means != NULL)
		free((*ctx)->means);

	free(*ctx);
	*ctx = NULL;
}

void
em_mog_init(struct em_mog_context *ctx)
{
	return_if_fail(ctx != NULL);
	int size, cov_size;
	size     = ctx->dim * ctx->ngauss;
	cov_size = ctx->dim * size;

	return_if_fail(size != 0);

	ctx->means      = malloc((size + ctx->ngauss + cov_size) * sizeof(double));
	return_if_fail(ctx->means != NULL);
	ctx->weight     = ctx->means + size;
	ctx->cov_matrix = ctx->weight + ctx->ngauss;
}

void
em_mog_clear(struct em_mog_context *ctx)
{
	return_if_fail(ctx != NULL);
	if (ctx->means != NULL)
		free(ctx->means);
	ctx->cov_matrix = NULL;
	ctx->means      = NULL;
	ctx->weight     = NULL;
	ctx->dim    = 0;
	ctx->ngauss = 0;
	ctx->epsilon = 0.0;
	ctx->min_loglike = 0.0;
}

/* Set error value for end of iteration process. */
void em_mog_set_error(struct em_mog_context *ctx, double error)
{
	return_val_if_fail(ctx != NULL, -1.0);
	ctx->epsilon = error;
}
/* Get error value for end of iteration process. */
double em_mog_get_error(struct em_mog_context *ctx)
{
	return_val_if_fail(ctx != NULL, -1.0);
	return ctx->epsilon;
}

/**
 * Set noice_level.
 */
void em_mog_set_noise_level(struct em_mog_context *ctx, double noise_level)
{
	return_if_fail(ctx != NULL);
	ctx->noise_level = noise_level;
}
/**
 * Get noise_level.
 */
double em_mog_get_noise_level(struct em_mog_context *ctx)
{
	return_val_if_fail(ctx != NULL, 0.0);
	return ctx->noise_level;
}

/* Set number of gaussians in mixture. */
void
em_mog_set_ngaussians(struct em_mog_context *ctx, int ngauss)
{
	return_if_fail(ctx != NULL);
	ctx->ngauss = ngauss;
}

/* Get number of gaussians in mixture. */
int
em_mog_get_ngaussians(struct em_mog_context *ctx)
{
	return_if_fail(ctx != NULL);
	return ctx->ngauss;
}

/* Set fitted data item dimension. */
void
em_mog_set_dimension(struct em_mog_context *ctx, int dim)
{
	return_if_fail(ctx != NULL);
	ctx->dim = dim;
}
/* Get fitted data item dimension. */
int
em_mog_get_dimension(struct em_mog_context *ctx)
{
	return_val_if_fail(ctx != NULL, EM_MOG_ERROR);
	return ctx->dim;
}

/* Set n-th gaussian component weight. */
void
em_mog_set_nth_weight(struct em_mog_context *ctx, int nth, double weight)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->weight != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);

	ctx->weight[nth] = weight;
}

/* Get n-th gaussian component weight. */
double
em_mog_get_nth_weight(struct em_mog_context *ctx, int nth)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->weight != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);

	return ctx->weight[nth];
}

/* Set n-th gaussian mean. */
void
em_mog_set_nth_mean(struct em_mog_context *ctx, int nth, const double *mu)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->means != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(mu != NULL);

	memcpy(&(ctx->means[nth * ctx->dim]), mu, sizeof(double) * ctx->dim);
}
/* Get n-th gaussian mean. */
void
em_mog_get_nth_mean(struct em_mog_context *ctx, int nth, double *mu)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->means != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(mu != NULL);

	memcpy(mu, &(ctx->means[nth * ctx->dim]), sizeof(double) * ctx->dim);
}

/* Set n-th gaussian covariance matrix. */
void
em_mog_set_nth_covariance_matrix(struct em_mog_context *ctx, int nth, const double *cmx)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->cov_matrix != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(cmx != NULL);
	int cov_size = ctx->dim * ctx->dim;

	memcpy(&(ctx->cov_matrix[nth * cov_size]), cmx, sizeof(double) * cov_size);
}
/* Get n-th gaussian covariance matrix. */
void
em_mog_get_nth_covariance_matrix(struct em_mog_context *ctx, int nth, double *cmx)
{
	return_if_fail(ctx != NULL);
	return_if_fail(ctx->cov_matrix != NULL);
	return_if_fail(nth > 0 && nth < ctx->ngauss);
	return_if_fail(cmx != NULL);
	int cov_size = ctx->dim * ctx->dim;

	memcpy(cmx, &(ctx->cov_matrix[nth * cov_size]), sizeof(double) * cov_size);
}

double
em_mog_step(struct em_mog_context *ctx, double *x, int nx)
{
	return_val_if_fail(ctx != NULL, 0.0);
	return_val_if_fail(ctx->means != NULL, 0.0);
	return_val_if_fail(nx > 0, 0.0);
	return_val_if_fail(x != NULL, 0.0);
	double loglike = 0.0;
	int res;

	double *probability = malloc(nx * ctx->ngauss * sizeof(double));
	return_val_if_fail(probability != NULL, loglike);

	res  = expectation_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	res = maximization_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	loglike = em_likelihood(ctx, x, nx);
quit:
	free(probability);
	return loglike;
}

int
em_mog_converge(struct em_mog_context *ctx, double *x, int nx, double epsilon)
{
	return_val_if_fail(ctx != NULL, 0.0);
	return_val_if_fail(ctx->means != NULL, 0.0);
	return_val_if_fail(nx > 0, 0.0);
	return_val_if_fail(x != NULL, 0.0);
	double err, loglike_prev, loglike = 0.0;
	int res;

	double *probability = malloc(nx * ctx->ngauss * sizeof(double));
	return_val_if_fail(probability != NULL, loglike);

	res  = expectation_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	res = maximization_step(probability, ctx, x, nx);
	if (res == EM_MOG_ERROR)
		goto quit;

	loglike_prev = em_likelihood(ctx, x, nx);

	do {
		res  = expectation_step(probability, ctx, x, nx);
		if (res == EM_MOG_ERROR)
			goto quit;

		res = maximization_step(probability, ctx, x, nx);
		if (res == EM_MOG_ERROR)
			goto quit;

		loglike = em_likelihood(ctx, x, nx);
		err = fabs(loglike - loglike_prev);
		loglike_prev = loglike;
	} while (err > epsilon);

#ifdef DEBUG
	int i, dim, cov_size;
	//printf("%d\n", count);
	dim = ctx->dim;
	cov_size = dim * dim;
	for (i = 0; i < ctx->ngauss; i++) {
		printf("%d\n", i + 1);
		printf_matrix(ctx->means + dim * i, dim, 1);
		printf("\n");
		printf("%d\n", i + 1);
		printf_matrix(ctx->cov_matrix + i * cov_size, dim, dim);
		printf("\n");
		printf("%d\n", i + 1);
		printf("%lf ", *(ctx->weight + i));
		printf("\n\n");
	}
#endif //DEBUG

quit:
	free(probability);
	return res;
}

int
em_mog_sample_to_match_model(struct em_mog_context *ctx, double *x)
{
	return_val_if_fail(ctx != NULL, EM_MOG_ERROR);
	return_val_if_fail(x != NULL, EM_MOG_ERROR);
	return_val_if_fail(ctx->means != NULL, EM_MOG_ERROR);

	int ret = EM_MOG_ERROR;
	int i, k, dim, cov_size;

	double *w, *m, *cmx, *p, tmp, dr, dg, db, mult;
	w   = ctx->weight;
	m   = ctx->means;
	cmx = ctx->cov_matrix;
	k   = ctx->ngauss;

	dim      = 3;
	cov_size = 9;
	mult = 12.2;
	/*
	 * (r_m - r) / (2.5 * sigma_r^2) + (g_m - g) / (2.5 * sigma_g^2)
	 * + (b_m - b) / (2.5 * sigma_b^2) <= 1
	 * rule two and a half sigma
	 */
	for (i = 0; i < k; i++) {
		p = x;

		tmp = *p - *m;
		dr = tmp * tmp;
		p++;
		m++;

		tmp = *p - *m;
		dg = tmp * tmp;
		p++;
		m++;

		tmp = *p - *m;
		db = tmp * tmp;

		ret = (dr < mult * cmx[0] && dg < mult * cmx[4] && db < mult * cmx[8]) ? i : EM_MOG_ERROR;
		if (ret == i)
			break;

		w++;
		cmx += cov_size;
	}

	return ret;
}

void
em_mog_update_nth_model(struct em_mog_context *ctx, double *data, int n_model)
{
	return_if_fail(ctx != NULL);
	return_if_fail(data != NULL);
	return_if_fail(ctx->means != NULL);

	double a1, a2, *m, *w, *cmx, *p;// vector[3];
	int dim, cmx_size;

	dim      = 3;
	cmx_size = 9;

	m   = ctx->means + n_model * dim;
	w   = ctx->weight + n_model;
	cmx = ctx->cov_matrix + n_model * cmx_size;
	p   = data;

	a1 = normal_distribution_diag_cov_mtrx(m, data, cmx);
	a2 = 1 - a1;

	//vector[0] = (*m - *p) * (*m - *p);
	*m      = a2 * *m + a1 * *p;
	m++;
	p++;
	//vector[1] = (*m - *p) * (*m - *p);
	*m      = a2 * *m + a1 * *p;
	m++;
	p++;
	//vector[2] = (*m - *p) * (*m - *p);
	*m      = a2 * *m + a1 * *p;

	/* TODO: матрица становится неположительно определённой*/
	/*p1 = vector;
	for (j = 0; j < dim; j++) {
		p2 = vector;
		for (l = 0; l < dim; l++) {
				if (i == j)
					*cmx = *cmx * a2 + a1 * (*p1) * (*p2);
			p2++;
			cmx++;
		}
		p1++;
	}*/
}

void em_mog_reset(struct em_mog_context *ctx, const double *data, int data_size)
{
	double *w, *m, *cov, weight;
	double cov_mtx[9] = {1.0, 0.0, 0.0,
						 0.0, 1.0, 0.0,
						 0.0, 0.0, 1.0};

	int i, k, dim, cov_size, r;
	k        = ctx->ngauss;
	dim      = 3;
	cov_size = 9;

	w   = ctx->weight;
	m   = ctx->means;
	cov = ctx->cov_matrix;

	weight = 1.0 / k;

	for (i = 0; i < k; i++) {
		*w = weight;
		memcpy(cov, cov_mtx, cov_size * sizeof(double));
		if (data != NULL && data_size > 0) {
			r = rand() % data_size;
			memcpy(m, data + r * dim, dim * sizeof(double));
		} else {
			memset(m, 0, dim * sizeof(double));
		}
		w++;
		m   += dim;
		cov += cov_size;
	}
}
